"""Route bundle with tapers to wider widths for reduced propagation loss."""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    mmi1 = c << gf.components.mmi1x2()
    mmi2 = c << gf.components.mmi1x2()
    mmi2.move((100, 50))
    route = gf.routing.route_bundle(
        c,
        [mmi1.ports["o2"]],
        [mmi2.ports["o1"]],
        taper=gf.components.taper(width1=0.5, width2=2),
        cross_section="strip",
        min_straight_taper=20,
    )
    c.show()



"""Route bundle electrical with enlarged bbox to force obstacle avoidance."""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    columns = 2
    ptop = c << gf.components.pad_array(columns=columns, port_orientation=270)
    pbot = c << gf.components.pad_array(port_orientation=270, columns=columns)
    ptop.movex(300)
    ptop.movey(300)

    obstacle = c << gf.c.rectangle(size=(300, 100), layer="M3", centered=True)
    obstacle.ymin = pbot.ymax - 10
    obstacle.xmin = pbot.xmax + 10

    routes = gf.routing.route_bundle_electrical(
        c,
        pbot.ports,
        ptop.ports,
        start_straight_length=100,
        separation=20,
        cross_section="metal_routing",
        bboxes=[
            obstacle.bbox().enlarge(10),
            pbot.bbox(),
            ptop.bbox(),
        ],
        sort_ports=True,
    )
    c.show()



from __future__ import annotations

__all__ = ["litho_steps"]

import gdsfactory as gf
from gdsfactory import components as pc
from gdsfactory.component import Component
from gdsfactory.typings import LayerSpec


@gf.cell_with_module_name
def litho_steps(
    line_widths: tuple[float, ...] = (1.0, 2.0, 4.0, 8.0, 16.0),
    line_spacing: float = 10.0,
    height: float = 100.0,
    layer: LayerSpec = "WG",
) -> Component:
    """Positive + negative tone linewidth test.

    used for lithography resolution test patterning
    based on phidl

    Args:
        line_widths: in um.
        line_spacing: in um.
        height: in um.
        layer: Specific layer to put the ruler geometry on.
    """
    D = gf.Component()

    height /= 2
    T1 = pc.text(
        text=f"{line_widths[-1]!s}", size=height, justify="center", layer=layer
    )

    ref = D.add_ref(T1)
    ref.rotate(90)
    ref.movex(-height / 10)

    R1 = pc.rectangle(size=(line_spacing, height), layer=layer)
    D.add_ref(R1).movey(-height)
    count = 0.0
    for i in reversed(line_widths):
        count += line_spacing + i
        R2 = pc.rectangle(size=(i, height), layer=layer)
        r = D.add_ref(R1)
        r.movex(count)
        r.movey(-height)
        r = D.add_ref(R2)
        r.movex(count - i)

    return D



"""Route bundle connecting ports facing the same direction (U-direct routing)."""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.port import Port


@gf.cell(cache={})
def test_connect_bundle_udirect(
    dy: int = 200, orientation: int = 270, layer: tuple[int, int] = (1, 0)
) -> gf.Component:
    xs1 = [-100, -90, -80, -55, -35, 24, 0] + [200, 210, 240]
    axis = "X" if orientation in [0, 180] else "Y"
    pitch = 10.0
    N = len(xs1)
    xs2 = [70 + i * pitch for i in range(N)]

    if axis == "X":
        ports1 = [
            Port(
                f"top_{i}",
                center=(0, xs1[i]),
                width=0.5,
                orientation=orientation,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports2 = [
            Port(
                f"bottom_{i}",
                center=(dy, xs2[i]),
                width=0.5,
                orientation=orientation,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
    else:
        ports1 = [
            Port(
                f"top_{i}",
                center=(xs1[i], 0),
                width=0.5,
                orientation=orientation,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports2 = [
            Port(
                f"bottom_{i}",
                center=(xs2[i], dy),
                width=0.5,
                orientation=orientation,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]

    c = gf.Component()
    gf.routing.route_bundle(
        c, ports1, ports2, radius=10.0, sort_ports=True, cross_section="strip"
    )
    return c


if __name__ == "__main__":
    c = test_connect_bundle_udirect()
    c.show()



"""Route bundle using all-angle (diagonal) routing."""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    rows = 3
    straight = gf.c.straight
    w1 = c << gf.c.array(straight, rows=rows, columns=1, row_pitch=10)
    w2 = c << gf.c.array(straight, rows=rows, columns=1, row_pitch=10)
    w2.rotate(-30)
    w2.movex(140)
    p1 = list(w1.ports.filter(orientation=0))
    p2 = list(w2.ports.filter(orientation=150))
    p1.reverse()
    p2.reverse()

    gf.routing.route_bundle_all_angle(
        c,
        p1,
        p2,
        separation=3,
    )
    c.show()



"""CD SEM structures."""

from __future__ import annotations

__all__ = ["cdsem_bend180"]

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import ComponentSpec, CrossSectionSpec

LINE_LENGTH = 420.0


@gf.cell_with_module_name
def cdsem_bend180(
    width: float = 0.5,
    radius: float = 10.0,
    wg_length: float | None = LINE_LENGTH,
    straight: ComponentSpec = "straight",
    bend90: ComponentSpec = "bend_circular",
    cross_section: CrossSectionSpec = "strip",
    text: ComponentSpec = "text_rectangular",
    text_size: float = 1.0,
) -> Component:
    """Returns CDSEM structures.

    Args:
        width: of the line.
        radius: um.
        wg_length: in um.
        straight: spec.
        bend90: spec.
        cross_section: spec.
        text: spec.
        text_size: um.
    """
    c = Component()
    r = radius

    if wg_length is None:
        wg_length = 2 * r

    bend90 = gf.get_component(
        bend90,
        cross_section=cross_section,
        radius=r,
        width=width,
        allow_min_radius_violation=True,
    )
    wg = gf.get_component(
        straight, cross_section=cross_section, length=wg_length, width=width
    )

    # Add the U-turn on straight layer
    b1 = c.add_ref(bend90)
    b2 = c.add_ref(bend90)
    b2.connect("o2", b1.ports["o1"])

    wg1 = c.add_ref(wg)
    wg1.connect("o1", b1.ports["o2"])

    wg2 = c.add_ref(wg)
    wg2.connect("o1", b2.ports["o1"])

    label = c << gf.get_component(text, text=str(int(width * 1e3)), size=text_size)
    label.ymax = b2.ymin - 5
    label.x = 0

    c2 = gf.Component()
    ref = c2 << c
    ref.rotate(90)
    c2.flatten()
    return c2



"""Route bundle all-angle connecting rotated mmi2x2 components."""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()

    mmi = gf.components.mmi2x2(width_mmi=10, gap_mmi=3)
    mmi1 = c.add_ref_off_grid(mmi)
    mmi2 = c.add_ref_off_grid(mmi)

    mmi2.move((100, 10))
    mmi2.rotate(30)

    routes = gf.routing.route_bundle_all_angle(
        c,
        mmi1.ports.filter(orientation=0),
        [mmi2.ports["o2"], mmi2.ports["o1"]],
    )
    c.show()



"""Lets try the bend_s_offset with different p parameters and see how the layout changes."""

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    with_arc_floorplan = True
    c1 = c << gf.c.bend_s_offset(
        offset=20, p=0.5, with_arc_floorplan=with_arc_floorplan
    )
    c2 = c << gf.c.bend_s_offset(
        offset=20, p=0.0, with_arc_floorplan=with_arc_floorplan
    )
    c3 = c << gf.c.bend_s_offset(
        offset=20, p=1.0, with_arc_floorplan=with_arc_floorplan
    )
    c.show()



"""Sample routing with steps."""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    c1 = c << gf.components.mmi2x2()
    c2 = c << gf.components.mmi2x2()
    c2.move((200, 240))

    gf.routing.route_bundle(
        c,
        [c1.ports["o3"], c1.ports["o4"]],
        [c2.ports["o2"], c2.ports["o1"]],
        steps=[{"dx": 60}, {"dy": 500}],
        separation=5.0,
        cross_section="strip",
        sort_ports=True,
        layer_marker=(1, 0),
    )
    c.show()



"""Route bundle with auto_taper for width mismatch between ports."""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    s1 = c << gf.components.straight()
    s2 = c << gf.components.straight(width=2)
    s2.move((40, 50))
    route = gf.routing.route_bundle(
        c,
        s1.ports["o2"],
        s2.ports["o1"],
        cross_section="strip",
        auto_taper=True,
    )
    c.show()



"""Sample route_bundle with non-Manhattan steps.

Steps that combine dx and dy in a single step produce a diagonal
(non-Manhattan) waypoint. The router automatically inserts corner
points to convert these into Manhattan-compatible segments.
"""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    c1 = c << gf.components.mmi2x2()
    c2 = c << gf.components.mmi2x2()
    c2.move((200, 240))

    # A single step with both dx and dy creates a non-Manhattan waypoint.
    # The router splits the diagonal into horizontal + vertical segments.
    gf.routing.route_bundle(
        c,
        [c1.ports["o3"], c1.ports["o4"]],
        [c2.ports["o2"], c2.ports["o1"]],
        steps=[{"dx": 60, "dy": 100}],
        separation=5.0,
        cross_section="strip",
        sort_ports=True,
    )
    c.show()



"""Route bundle connecting top and bottom ports vertically."""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    xs_top = [0, 10, 20, 40, 50, 80]
    pitch = 127.0
    N = len(xs_top)
    xs_bottom = [(i - N / 2) * pitch for i in range(N)]
    layer = (1, 0)

    top_ports = [
        gf.Port(
            f"top_{i}",
            center=(xs_top[i], 0),
            width=0.5,
            orientation=270,
            layer=gf.get_layer(layer),
        )
        for i in range(N)
    ]

    bot_ports = [
        gf.Port(
            f"bot_{i}",
            center=(xs_bottom[i], -300),
            width=0.5,
            orientation=90,
            layer=gf.get_layer(layer),
        )
        for i in range(N)
    ]

    c = gf.Component()
    routes = gf.routing.route_bundle(
        c,
        top_ports,
        bot_ports,
        separation=5.0,
        end_straight_length=100,
        cross_section="strip",
    )
    c.show()



import gdsfactory as gf

if __name__ == "__main__":
    s0 = gf.Section(
        width=1.2, offset=0, layer=(2, 0), name="core", port_names=("o1", "o2")
    )
    s1 = gf.Section(width=2.2, offset=0, layer=(3, 0), name="etch")
    s2 = gf.Section(width=1.1, offset=3, layer=(1, 0), name="wg2")
    X1 = gf.CrossSection(sections=(s0, s1, s2))

    # Create the second Cross-section that we want to transition to.
    s0 = gf.Section(
        width=1, offset=0, layer=(2, 0), name="core", port_names=("o1", "o2")
    )
    s1 = gf.Section(width=3.5, offset=0, layer=(3, 0), name="etch")
    s2 = gf.Section(width=3, offset=5, layer=(1, 0), name="wg2")
    X2 = gf.CrossSection(sections=(s0, s1, s2))

    # To show the cross-sections, let us now create two paths and create components by extruding them.
    P1 = gf.path.straight(length=5)
    P2 = gf.path.straight(length=5)
    wg1 = gf.path.extrude(P1, X1, register_cross_section=True)
    wg2 = gf.path.extrude(P2, X2, register_cross_section=True)

    # Place both cross-section components and quickplot them,
    # Quickplot is designed to create a wide variety of complex graphs with a simple, concise syntax, making it ideal for quick data exploration.
    c = gf.Component()
    wg1ref = c << wg1
    wg2ref = c << wg2
    wg2ref.movex(7.5)

    # Create the transitional cross-section.
    # Xtrans = gf.path.transition(cross_section1=X1, cross_section2=X2, width_type="sine")
    Xtrans = gf.path.transition(
        cross_section1=wg1["o2"].info["cross_section"],
        cross_section2=wg2["o1"].info["cross_section"],
        width_type="sine",
    )

    # Create a Path for the transitional cross-section to follow.
    P3 = gf.path.straight(length=15, npoints=100)

    # Use the transitional cross-section to create a component.
    c = gf.path.extrude_transition(P3, Xtrans)

    c.show()



from __future__ import annotations

from collections.abc import Sequence
from functools import partial

import numpy as np
from kfactory.routing.generic import ManhattanRoute

import gdsfactory as gf
from gdsfactory import typings
from gdsfactory.component import Component, ComponentReference
from gdsfactory.cross_section import CrossSection
from gdsfactory.port import Port, select_ports_optical
from gdsfactory.routing.auto_taper import add_auto_tapers
from gdsfactory.routing.route_bundle import route_bundle
from gdsfactory.routing.utils import direction_ports_from_list_ports
from gdsfactory.typings import (
    ComponentSpec,
    CrossSectionSpec,
    SelectPorts,
    Strs,
)


def route_south(
    component: Component,
    component_to_route: Component | ComponentReference,
    optical_routing_type: int = 1,
    excluded_ports: Sequence[str] | None = None,
    straight_separation: float = 4.0,
    io_gratings_lines: list[list[ComponentReference]] | None = None,
    gc_port_name: str = "o1",
    bend: ComponentSpec = "bend_euler",
    straight: ComponentSpec = "straight",
    select_ports: SelectPorts = select_ports_optical,
    port_names: Strs | None = None,
    cross_section: CrossSectionSpec = "strip",
    start_straight_length: float = 0.5,
    port_type: str | None = None,
    allow_width_mismatch: bool = False,
    auto_taper: bool = True,
) -> list[ManhattanRoute]:
    """Places routes to route a component ports to the south.

    Args:
        component: top level component to add the routes.
        component_to_route: component or reference to route ports to south.
        optical_routing_type: routing heuristic `1` or `2` \
            1: uses the component size info to estimate the box size.\
            2: only looks at the optical port positions to estimate the size.
        excluded_ports: list of port names to NOT route.
        straight_separation: in um.
        io_gratings_lines: list of ports to which the ports produced by this function will be connected. \
                Supplying this information helps avoiding straight collisions.
        gc_port_name: grating coupler port name. Used only if io_gratings_lines is supplied.
        bend: spec.
        straight: spec.
        select_ports: function to select_ports.
        port_names: optional port names. Overrides select_ports.
        cross_section: cross_section spec.
        start_straight_length: in um.
        port_type: optical or electrical.
        allow_width_mismatch: allow width mismatch.
        auto_taper: auto taper.

    Works well if the component looks roughly like a rectangular box with:
        north ports on the north of the box.
        south ports on the south of the box.
        east ports on the east of the box.
        west ports on the west of the box.

    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component()
        ref = c << gf.components.ring_double()
        r = gf.routing.route_south(c, ref)
        c.plot()

    """
    xs = gf.get_cross_section(cross_section)
    excluded_ports = excluded_ports or ()
    start_straight_length0 = start_straight_length
    routes: list[ManhattanRoute] = []

    if optical_routing_type not in {1, 2}:
        raise ValueError(
            f"optical_routing_type = {optical_routing_type} not in supported [1, 2]"
        )

    if port_names:
        optical_ports = [component_to_route[port_name] for port_name in port_names]
    else:
        optical_ports = [
            p
            for p in select_ports(component_to_route.ports)
            if p.name not in excluded_ports
        ]

    if auto_taper:
        optical_ports = add_auto_tapers(component, optical_ports, cross_section)

    if not optical_ports:
        return []

    port_type = port_type or optical_ports[0].port_type
    bend90 = bend(cross_section=cross_section) if callable(bend) else bend
    bend90 = gf.get_component(bend90)
    dy = abs(bend90.info["dy"])

    # Handle empty list gracefully

    route_with_conn_params = partial(
        route_bundle,
        bend=bend,
        straight=straight,
        cross_section=cross_section,
        port_type=port_type,
        allow_width_mismatch=allow_width_mismatch,
        auto_taper=False,
    )

    # Used to avoid crossing between straights in special cases
    # This could happen when abs(x_port - x_grating) <= 2 * dy
    delta_gr_min = 2 * dy + 1
    sep = straight_separation

    # Get lists of optical ports by orientation
    direction_ports = direction_ports_from_list_ports(optical_ports)

    north_ports = direction_ports["N"]
    north_start = north_ports[: len(north_ports) // 2]
    north_finish = north_ports[len(north_ports) // 2 :]

    west_ports = direction_ports["W"]
    west_ports.reverse()
    east_ports = direction_ports["E"]
    south_ports = direction_ports["S"]
    north_finish.reverse()  # Sort right to left
    north_start.reverse()  # Sort right to left
    ordered_ports = north_start + west_ports + south_ports + east_ports + north_finish

    def get_index_port_closest_to_x(
        x: float, component_references: list[ComponentReference]
    ) -> np.intp:
        return np.array(
            [abs(x - p.ports[gc_port_name].x) for p in component_references]
        ).argmin()

    def gen_port_from_port(
        x: float, y: float, p: typings.Port, cross_section: CrossSection
    ) -> Port:
        return Port(
            name=p.name,
            center=(x, y),
            orientation=90.0,
            width=p.width,
            layer=gf.get_layer(cross_section.layer),
            port_type=p.port_type,
        )

    west_ports.reverse()
    y0 = min(p.y for p in ordered_ports) - dy - 0.5
    ports_to_route: list[Port] = []

    optical_xs_tmp = [p.x for p in ordered_ports]
    x_optical_min = min(optical_xs_tmp)
    x_optical_max = max(optical_xs_tmp)

    # Set starting ``x`` on the west side
    # ``x`` is the x-coord of the waypoint where the current component port is connected.
    # x starts as close as possible to the component.
    # For each new port, the distance is increased by the separation.
    # The starting x depends on the heuristic chosen : ``1`` or ``2``
    if optical_routing_type == 1:
        # use component size to know how far to route
        x = component_to_route.xmin - dy - 1
    else:
        # use optical port to know how far to route
        x = x_optical_min - dy - 1

    # First route the ports facing west
    # In case we have to connect these ports to a line of gratings,
    # Ensure that the port is aligned with the grating port or
    # has enough space for manhattan routing (at least two bend radius)
    for p in west_ports:
        if io_gratings_lines:
            i_grating = get_index_port_closest_to_x(x, io_gratings_lines[-1])
            x_gr = io_gratings_lines[-1][i_grating].ports[gc_port_name].x
            if abs(x - x_gr) < delta_gr_min:
                if x > x_gr:
                    x = x_gr
                elif x < x_gr:
                    x = x_gr - delta_gr_min

        tmp_port = gen_port_from_port(x, y0, p, cross_section=xs)
        ports_to_route.append(tmp_port)
        route = route_with_conn_params(component, tmp_port, p)[0]
        x -= sep

    # route first halft of north ports above the top west one
    north_start.reverse()  # We need them from left to right

    start_straight_length = start_straight_length0
    if len(north_start) > 0:
        y_max = max(p.y for p in west_ports + north_start)
        for p in north_start:
            tmp_port = gen_port_from_port(x, y0, p, cross_section=xs)
            route = route_with_conn_params(
                component,
                tmp_port,
                p,
                end_straight_length=start_straight_length + y_max - p.y,
            )[0]

            ports_to_route.append(tmp_port)
            x -= sep
            start_straight_length += sep
            routes.append(route)

    # Set starting ``x`` on the east side
    if optical_routing_type == 1:
        #  use component size to know how far to route
        x = component_to_route.xmax + dy + 1
    elif optical_routing_type == 2:
        # use optical port to know how far to route
        x = x_optical_max + dy + 1

    # Route the east ports
    # In case we have to connect these ports to a line of gratings,
    # Ensure that the port is aligned with the grating port or
    # has enough space for manhattan routing (at least two bend radius)
    start_straight_length = start_straight_length0
    for p in east_ports:
        if io_gratings_lines:
            i_grating = get_index_port_closest_to_x(x, io_gratings_lines[-1])
            x_gr = io_gratings_lines[-1][i_grating].ports[gc_port_name].x
            if abs(x - x_gr) < delta_gr_min:
                if x < x_gr:
                    x = x_gr
                elif x > x_gr:
                    x = x_gr + delta_gr_min

        tmp_port = gen_port_from_port(x, y0, p, cross_section=xs)
        route = route_with_conn_params(
            component,
            tmp_port,
            p,
            end_straight_length=start_straight_length,
        )[0]
        routes.append(route)
        ports_to_route.append(tmp_port)
        x += sep

    # Route the remaining north ports
    start_straight_length = start_straight_length0
    if len(north_finish) > 0:
        y_max = max(p.y for p in east_ports + north_finish)
        for p in north_finish:
            tmp_port = gen_port_from_port(x, y0, p, cross_section=xs)
            ports_to_route.append(tmp_port)
            route = route_with_conn_params(
                component,
                tmp_port,
                p,
                end_straight_length=start_straight_length + y_max - p.y,
            )[0]
            x += sep
            start_straight_length += sep
            routes.append(route)

    flipped_ports = [p.copy() for p in ports_to_route]
    for p in flipped_ports:
        p.trans *= gf.kdb.Trans.R180
    component.add_ports(flipped_ports)
    component.add_ports(south_ports)
    component.auto_rename_ports()
    return routes



"""Register a CrossSection factory when adding a port.

When adding a port with a cross_section, you can set
``register_cross_section=True`` so the CrossSection is registered
in the active PDK.  This lets you later retrieve it by name with
``gf.get_cross_section``.
"""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component


@gf.cell
def component_with_registered_cross_section(
    length: float = 5.0,
    width: float = 2.5,
    layer: gf.typings.LayerSpec = "WG",
) -> Component:
    """Returns a component whose port cross_section is registered in the PDK.

    Args:
        length: in um.
        width: waveguide width in um.
        layer: layer.
    """
    xs = gf.cross_section.pin(width=width, layer=layer)

    c = gf.Component()
    p = gf.path.straight(length=length)
    c = p.extrude(
        xs, register_cross_section=True
    )  # extrude the path to create a component with a polygon geometry

    # c.add_polygon([(0, 0), (length, 0), (length, width), (0, width)], layer=layer)
    # c.add_port(
    #     name="o1",
    #     center=(0, width / 2),
    #     orientation=180,
    #     cross_section=xs,
    #     register_cross_section=True,
    # )

    # The cross_section is now retrievable from the PDK by name.
    xs_name = c["o1"].info["cross_section"]
    assert gf.get_cross_section(xs_name) == xs
    # c = gf.c.extend_ports(c, cross_section=c['o1'].info['cross_section'])
    c = gf.c.extend_ports(c, cross_section=xs)
    return c


def test_component_with_registered_cross_section() -> None:
    assert component_with_registered_cross_section()


if __name__ == "__main__":
    c = component_with_registered_cross_section()
    c.show()



"""Route bundle connecting two mmi2x2 components."""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    c1 = c << gf.components.mmi2x2()
    c2 = c << gf.components.mmi2x2()

    c2.move((100, 50))
    routes = gf.routing.route_bundle(
        c,
        [c1.ports["o4"], c1.ports["o3"]],
        [c2.ports["o1"], c2.ports["o2"]],
        radius=5,
        cross_section="strip",
    )
    c.show()



"""Write a GDS with all cells."""

import gdsfactory as gf
from gdsfactory.gpdk import PDK

cells = gf.c

if __name__ == "__main__":
    PDK.activate()

    c0 = cells.straight_heater_meander()
    c1 = cells.via_stack_m1_mtop()
    c2 = cells.via_stack_heater_mtop()
    c = gf.grid([c0, c1, c2])
    c.show()
    # s = c.to_3d()
    # s.show()



"""Route using Dubins paths for shortest path with minimal bending."""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()

    wg1 = c << gf.components.straight(length=100, width=3.2)
    wg2 = c << gf.components.straight(length=100, width=3.2)

    wg2.move((300, 50))
    wg2.rotate(45)

    route = gf.routing.route_dubins(
        c,
        port1=wg1.ports["o2"],
        port2=wg2.ports["o1"],
        cross_section=gf.cross_section.strip(width=3.2, radius=100),
    )
    c.show()



from __future__ import annotations

__all__ = ["ring"]

import numpy as np
from numpy import cos, pi, sin

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import LayerSpec


@gf.cell_with_module_name
def ring(
    radius: float = 10.0,
    width: float = 0.5,
    angle_resolution: float = 2.5,
    layer: LayerSpec = "WG",
    angle: float = 360,
    distance_resolution: float | None = None,
) -> Component:
    """Returns a ring.

    Args:
        radius: ring radius.
        width: of the ring.
        angle_resolution: max number of degrees per point.
        layer: layer.
        angle: angular coverage of the ring
        distance_resolution: max distance between points. This is an alternate way to describe the resolution besides setting angle_resolution. If distance_resolution and angle_resolution are both set, distance_resolution determines the resolution.
    """
    if radius < width / 2:
        raise ValueError(
            f"Error: radius is {radius} and width is {width}. radius must be >= width / 2."
        )

    if width < 0:
        raise ValueError(f"Error: width is {width}, but it must be nonnegative.")

    if angle > 360 or angle < 0:
        raise ValueError(f"Error: angle is {angle}, but it must be in [0, 360].")

    if distance_resolution is not None and distance_resolution <= 0:
        raise ValueError(
            f"Error: distance_resolution is {distance_resolution}, but it must be positive if given."
        )

    if distance_resolution is None and angle_resolution <= 0:
        raise ValueError(
            f"Error: angle_resolution is {angle_resolution}, but it must be positive."
        )

    D = gf.Component()
    inner_radius = radius - width / 2
    outer_radius = radius + width / 2
    if distance_resolution is not None:
        num_points = int(
            np.ceil(2 * pi * outer_radius * angle / 360 / distance_resolution)
        )
    else:
        num_points = int(np.ceil(angle / angle_resolution))
    t = np.linspace(0, angle, num_points + 1) * pi / 180
    inner_points_x = inner_radius * cos(t)
    inner_points_y = inner_radius * sin(t)
    outer_points_x = outer_radius * cos(t)
    outer_points_y = outer_radius * sin(t)
    xpts = np.concatenate([inner_points_x, outer_points_x[::-1]])
    ypts = np.concatenate([inner_points_y, outer_points_y[::-1]])
    D.add_polygon(points=list(zip(xpts, ypts, strict=False)), layer=layer)
    return D



"""Sample route_bundle with absolute x/y steps.

- x: sets the absolute x coordinate of the next waypoint.
- y: sets the absolute y coordinate of the next waypoint.
"""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    mmi1 = c << gf.components.mmi1x2()
    mmi2 = c << gf.components.mmi1x2()
    mmi2.move((200, 50))

    # x=100 sets the waypoint x-coordinate to 100 (absolute)
    gf.routing.route_single(
        c,
        mmi1.ports["o2"],
        mmi2.ports["o1"],
        steps=[{"dx": 100}],
        cross_section="strip",
        # layer_marker=(1,0)
    )
    c.show()



if __name__ == "__main__":
    import gdsfactory as gf

    # gf.CONF.pdk = 'gdsfactory.gpdk'

    pitch = 127.0
    xs_top = [-100, -90, -80, 0, 10, 20, 40, 50, 80, 90, 100, 105, 110, 115]
    N = len(xs_top)
    xs_bottom = [(i - N / 2) * pitch for i in range(N)]
    layer = 1

    top_ports = [
        gf.Port(
            name=f"top_{i}",
            center=(xs_top[i], 0),
            width=0.5,
            orientation=270,
            layer=layer,
        )
        for i in range(N)
    ]

    bot_ports = [
        gf.Port(
            name=f"bot_{i}",
            center=(xs_bottom[i], -400),
            width=0.5,
            orientation=90,
            layer=layer,
        )
        for i in range(N)
    ]

    c = gf.Component(name="test_route_bundle")
    routes = gf.routing.route_bundle(
        c,
        top_ports,
        bot_ports,
        start_straight_length=5,
        end_straight_length=10,
        cross_section="strip",
    )
    lengths = {i: route.length for i, route in enumerate(routes)}



"""Route bundle through corners connecting horizontal and vertical port groups."""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.port import Port


@gf.cell(cache={})
def test_connect_corner(N: int = 6, config: str = "A") -> gf.Component:
    d = 10.0
    sep = 5.0
    c = gf.Component()
    layer = (1, 0)

    if config in ["A", "B"]:
        a = 100.0
        ports_A_TR = [
            Port(
                f"A_TR_{i}",
                center=(d, a / 2 + i * sep),
                width=0.5,
                orientation=0,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_A_TL = [
            Port(
                f"A_TL_{i}",
                center=(-d, a / 2 + i * sep),
                width=0.5,
                orientation=180,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_A_BR = [
            Port(
                f"A_BR_{i}",
                center=(d, -a / 2 - i * sep),
                width=0.5,
                orientation=0,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_A_BL = [
            Port(
                f"A_BL_{i}",
                center=(-d, -a / 2 - i * sep),
                width=0.5,
                orientation=180,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_A = [ports_A_TR, ports_A_TL, ports_A_BR, ports_A_BL]

        ports_B_TR = [
            Port(
                f"B_TR_{i}",
                center=(a / 2 + i * sep, d),
                width=0.5,
                orientation=90,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_B_TL = [
            Port(
                f"B_TL_{i}",
                center=(-a / 2 - i * sep, d),
                width=0.5,
                orientation=90,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_B_BR = [
            Port(
                f"B_BR_{i}",
                center=(a / 2 + i * sep, -d),
                width=0.5,
                orientation=270,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_B_BL = [
            Port(
                f"B_BL_{i}",
                center=(-a / 2 - i * sep, -d),
                width=0.5,
                orientation=270,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_B = [ports_B_TR, ports_B_TL, ports_B_BR, ports_B_BL]

    elif config in ["C", "D"]:
        a = N * sep + 2 * d
        ports_A_TR = [
            Port(
                f"A_TR_{i}",
                center=(a, d + i * sep),
                width=0.5,
                orientation=0,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_A_TL = [
            Port(
                f"A_TL_{i}",
                center=(-a, d + i * sep),
                width=0.5,
                orientation=180,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_A_BR = [
            Port(
                f"A_BR_{i}",
                center=(a, -d - i * sep),
                width=0.5,
                orientation=0,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_A_BL = [
            Port(
                f"A_BL_{i}",
                center=(-a, -d - i * sep),
                width=0.5,
                orientation=180,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_A = [ports_A_TR, ports_A_TL, ports_A_BR, ports_A_BL]

        ports_B_TR = [
            Port(
                f"B_TR_{i}",
                center=(d + i * sep, a),
                width=0.5,
                orientation=90,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_B_TL = [
            Port(
                f"B_TL_{i}",
                center=(-d - i * sep, a),
                width=0.5,
                orientation=90,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_B_BR = [
            Port(
                f"B_BR_{i}",
                center=(d + i * sep, -a),
                width=0.5,
                orientation=270,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_B_BL = [
            Port(
                f"B_BL_{i}",
                center=(-d - i * sep, -a),
                width=0.5,
                orientation=270,
                layer=gf.get_layer(layer),
            )
            for i in range(N)
        ]
        ports_B = [ports_B_TR, ports_B_TL, ports_B_BR, ports_B_BL]

    if config in ["A", "C"]:
        for ports1, ports2 in zip(ports_A, ports_B, strict=True):
            gf.routing.route_bundle(
                c, ports1, ports2, radius=5, sort_ports=True, cross_section="strip"
            )
    elif config in ["B", "D"]:
        for ports1, ports2 in zip(ports_A, ports_B, strict=True):
            gf.routing.route_bundle(
                c, ports2, ports1, radius=5, sort_ports=True, cross_section="strip"
            )

    return c


if __name__ == "__main__":
    c = test_connect_corner(config="A")
    c.show()



"""Sample route_bundle with steps using dx/dy (relative) and x/y (absolute).

- dx/dy: shift relative to the current position.
- x/y: set an absolute coordinate.
"""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    mmi1 = c << gf.components.mmi1x2()
    mmi2 = c << gf.components.mmi1x2()
    mmi2.move((200, 50))

    # dx shifts the route 60um to the right from the current position
    gf.routing.route_bundle(
        c,
        mmi1.ports["o2"],
        mmi2.ports["o1"],
        steps=[{"dx": 60}],
        cross_section="strip",
    )
    c.show()



"""Routes bundles of ports (river routing).

get bundle is the generic river routing function
route_bundle calls different function depending on the port orientation.

 - route_bundle_same_axis: ports facing each other with arbitrary pitch on each side
 - route_bundle_corner: 90Deg / 270Deg between ports with arbitrary pitch
 - route_bundle_udirect: ports with direct U-turns
 - route_bundle_uindirect: ports with indirect U-turns

"""

from __future__ import annotations

import warnings
from collections.abc import Sequence
from functools import partial
from typing import Any, Literal, cast
from warnings import warn

import kfactory as kf
from kfactory.routing.generic import ManhattanRoute
from kfactory.routing.optical import PathLengthConfig

import gdsfactory as gf
from gdsfactory.routing.auto_taper import add_auto_tapers
from gdsfactory.routing.sort_ports import get_port_x, get_port_y
from gdsfactory.typings import (
    STEP_DIRECTIVES,
    ComponentSpec,
    Coordinates,
    CrossSectionSpec,
    LayerSpec,
    LayerSpecs,
    LayerTransitions,
    Port,
    Ports,
    Step,
)

OpticalManhattanRoute = ManhattanRoute

TOLERANCE = 1


def get_min_spacing(
    ports1: Ports,
    ports2: Ports,
    separation: float = 5.0,
    radius: float = 5.0,
    sort_ports: bool = True,
) -> float:
    """Returns the minimum amount of spacing in um required to create a fanout.

    Args:
        ports1: first list of ports.
        ports2: second list of ports.
        separation: minimum separation between two straights in um.
        radius: bend radius in um.
        sort_ports: sort the ports according to the axis.

    """
    if not ports1 or not ports2:
        raise ValueError("ports1 and ports2 must be non-empty")
    if len(ports1) != len(ports2):
        raise ValueError(f"ports1={len(ports1)} and ports2={len(ports2)} must be equal")

    axis = "X" if ports1[0].orientation in [0, 180] else "Y"
    j = 0
    min_j = 0
    max_j = 0
    if sort_ports:
        if axis in {"X", "x"}:
            ports1 = sorted(ports1, key=get_port_y)
            ports2 = sorted(ports2, key=get_port_y)
        else:
            ports1 = sorted(ports1, key=get_port_x)
            ports2 = sorted(ports2, key=get_port_x)

    for port1, port2 in zip(ports1, ports2, strict=False):
        if axis in {"X", "x"}:
            x1 = get_port_y(port1)
            x2 = get_port_y(port2)
        else:
            x1 = get_port_x(port1)
            x2 = get_port_x(port2)
        if x2 >= x1:
            j += 1
        else:
            j -= 1
        if j < min_j:
            min_j = j
        if j > max_j:
            max_j = j
    return (max_j - min_j) * separation + 2 * radius + 1.0


def _ensure_manhattan_waypoints(
    waypoints: list[kf.kdb.DPoint],
    start_port: gf.Port | None = None,
) -> list[kf.kdb.DPoint]:
    """Insert corner points between non-Manhattan waypoints to make the path Manhattan.

    For each pair of consecutive waypoints that are not axis-aligned,
    an intermediate corner point is inserted so all segments are
    either purely horizontal or purely vertical.

    Args:
        waypoints: list of waypoints that may contain non-Manhattan segments.
        start_port: optional start port to determine initial routing direction.

    Returns:
        list of waypoints with corner points inserted where needed.
    """
    if len(waypoints) < 2:
        return list(waypoints)

    tol = 1e-3
    result = [waypoints[0]]

    for i in range(1, len(waypoints)):
        prev = result[-1]
        curr = waypoints[i]

        dx = abs(curr.x - prev.x)
        dy = abs(curr.y - prev.y)

        if dx < tol or dy < tol:
            result.append(curr)
            continue

        # Non-Manhattan segment - insert a corner point
        if len(result) >= 2:
            prev_prev = result[-2]
            last_horizontal = abs(prev.x - prev_prev.x) > abs(prev.y - prev_prev.y)
            go_horizontal_first = not last_horizontal
        elif start_port is not None and start_port.orientation is not None:
            go_horizontal_first = int(start_port.orientation) % 360 in {0, 180}
        else:
            go_horizontal_first = True

        if go_horizontal_first:
            result.append(kf.kdb.DPoint(curr.x, prev.y))
        else:
            result.append(kf.kdb.DPoint(prev.x, curr.y))

        result.append(curr)

    return result


def route_bundle(
    component: gf.Component,
    ports1: Port | Ports | None = None,
    ports2: Port | Ports | None = None,
    cross_section: CrossSectionSpec | None = None,
    layer: LayerSpec | None = None,
    separation: float = 3.0,
    bend: ComponentSpec = "bend_euler",
    sort_ports: bool = False,
    start_straight_length: float = 0,
    end_straight_length: float = 0,
    min_straight_taper: float = 100,
    taper: ComponentSpec | None = None,
    port_type: str | None = None,
    collision_check_layers: LayerSpecs | None = None,
    on_collision: Literal["error", "show_error"] | None = None,
    bboxes: Sequence[kf.kdb.DBox] | None = None,
    allow_width_mismatch: bool | None = None,
    allow_layer_mismatch: bool | None = None,
    allow_type_mismatch: bool | None = None,
    radius: float | None = None,
    route_width: float | None = None,
    straight: ComponentSpec = "straight",
    sbend: ComponentSpec | None = None,
    auto_taper: bool = True,
    auto_taper_taper: ComponentSpec | None = None,
    waypoints: Coordinates | Sequence[gf.kdb.DPoint] | None = None,
    steps: Sequence[Step] | None = None,
    start_angles: float | list[float] | None = None,
    end_angles: float | list[float] | None = None,
    router: Literal["optical", "electrical"] | None = None,
    layer_transitions: LayerTransitions | None = None,
    show_waypoints: bool = False,
    layer_marker: LayerSpec | None = None,
    raise_on_error: bool = False,
    path_length_matching_config: PathLengthConfig | None = None,
    layer_label: LayerSpec | None = None,
    port1: Port | None = None,
    port2: Port | None = None,
) -> list[ManhattanRoute]:
    """Places a bundle of routes to connect two groups of ports.

    Routes connect a bundle of ports with a river router.
    Chooses the correct routing function depending on port angles.

    Can also be used with single ports instead of lists, replacing route_bundle.

    Args:
        component: component to add the routes to.
        ports1: starting port or list of starting ports.
        ports2: end port or list of end ports.
        cross_section: CrossSection or function that returns a cross_section.
        layer: layer to use for the route.
        separation: bundle separation (center to center). Defaults to cross_section.width + cross_section.gap
        bend: function for the bend. Defaults to euler.
        sort_ports: sort port coordinates.
        start_straight_length: straight length at the beginning of the route. If None, uses default value for the routing CrossSection.
        end_straight_length: end length at the beginning of the route. If None, uses default value for the routing CrossSection.
        min_straight_taper: minimum length for tapering the straight sections.
        taper: function for tapering long straight waveguides beyond min_straight_taper. Defaults to None.
        port_type: type of port to place. Defaults to optical.
        collision_check_layers: list of layers to check for collisions.
        on_collision: action to take on collision. Defaults to None (ignore).
        bboxes: list of bounding boxes to avoid collisions.
        allow_width_mismatch: allow different port widths.
        allow_layer_mismatch: allow different port layers to connect.
        allow_type_mismatch: allow different port types to connect.
        radius: bend radius. If None, defaults to cross_section.radius.
        route_width: width of the route. If None, defaults to cross_section.width.
        straight: function for the straight. Defaults to straight.
        sbend: function for the s-bend. If None, uses the same function as bend.
        auto_taper: if True, auto-tapers ports to the cross-section of the route.
        auto_taper_taper: taper to use for auto-tapering. If None, uses the default taper for the cross-section.
        waypoints: list of waypoints to add to the route.
        steps: list of steps to add to the route.
            Each step is a dict with keys: x (absolute), y (absolute), dx (relative), dy (relative).
            Use x/y to set an absolute coordinate and dx/dy to shift relative to the current position.
        start_angles: list of start angles for the routes. Only used for electrical ports.
        end_angles: list of end angles for the routes. Only used for electrical ports.
        router: Set the type of router to use, either the optical one or the electrical one.
            If None, the router is optical unless the port_type is "electrical".
        layer_transitions: dictionary of layer transitions to use for the routing when auto_taper=True.
        show_waypoints: if True, places markers at each waypoint using CONF.layer_marker.
        layer_marker: layer to place markers on the route. Overrides CONF.layer_marker when show_waypoints=True.
        raise_on_error: if True, raises an exception on routing error instead of adding error markers.
        layer_label: layer to place length labels on the route.

    .. plot::
        :include-source:

        import gdsfactory as gf

        dy = 200.0
        xs1 = [-500, -300, -100, -90, -80, -55, -35, 200, 210, 240, 500, 650]

        pitch = 10.0
        N = len(xs1)
        xs2 = [-20 + i * pitch for i in range(N // 2)]
        xs2 += [400 + i * pitch for i in range(N // 2)]

        a1 = 90
        a2 = a1 + 180

        ports1 = [gf.Port(name=f"top_{i}", center=(xs1[i], +0), width=0.5, orientation=a1, layer=(1, 0)) for i in range(N)]
        ports2 = [gf.Port(name=f"bot_{i}", center=(xs2[i], dy), width=0.5, orientation=a2, layer=(1, 0)) for i in range(N)]

        c = gf.Component()
        gf.routing.route_bundle(component=c, ports1=ports1, ports2=ports2, cross_section='strip', separation=5)
        c.plot()
    """
    # Support deprecated port1/port2 keyword arguments
    if port1 is not None:
        if ports1 is not None:
            raise ValueError("Cannot specify both ports1 and port1")
        ports1 = port1
    if port2 is not None:
        if ports2 is not None:
            raise ValueError("Cannot specify both ports2 and port2")
        ports2 = port2

    if ports1 is None or ports2 is None:
        raise ValueError("ports1 and ports2 are required")

    # Wrap single ports in lists
    if isinstance(ports1, kf.DPort):
        ports1 = [ports1]
    if isinstance(ports2, kf.DPort):
        ports2 = [ports2]

    if show_waypoints and layer_marker is None:
        layer_marker = gf.CONF.layer_marker

    component = gf.Component(base=component.base)  # type: ignore[call-overload]
    ports1 = [gf.Port(base=p1.base) for p1 in ports1]
    ports2 = [gf.Port(base=p2.base) for p2 in ports2]

    if router:
        warnings.warn(
            f"The argument {router=} is ignored and will be removed in a future release.",
            stacklevel=2,
        )

    router = "optical"

    if cross_section is None:
        if layer is None or route_width is None:
            raise ValueError(
                f"Either {cross_section=} or {layer=} and {route_width=} must be provided"
            )
    elif layer is not None:
        raise ValueError(
            f"Cannot have both {layer=} and {cross_section=} provided. Choose one."
        )

    c = component
    ports1_ = list(ports1)
    ports2_ = list(ports2)
    port_type = port_type or ports1_[0].port_type

    if cross_section is None:
        cross_section = partial(
            gf.cross_section.cross_section,
            layer=cast("LayerSpec", layer),
            width=cast("float", route_width),
            port_names=("e1", "e2") if port_type == "electrical" else ("o1", "o2"),
            port_types=(port_type, port_type),
        )

    if len(ports1_) != len(ports2_):
        raise ValueError(
            f"ports1={len(ports1_)} and ports2={len(ports2_)} must be equal"
        )
    if route_width:
        xs = gf.get_cross_section(cross_section, width=route_width)
    else:
        xs = gf.get_cross_section(cross_section)
    width = route_width or xs.width

    radius = radius or xs.radius
    taper_cell = gf.get_component(taper) if taper else None

    if collision_check_layers:
        collision_check_layer_enums = [
            gf.get_layer(layer) for layer in collision_check_layers
        ]
    else:
        collision_check_layer_enums = None

    bboxes = list(bboxes or [])

    if auto_taper and auto_taper_taper:
        warn(
            "Use of `auto_taper_taper` is deprecated. Please use `layer_transitions` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        taper_ = gf.get_component(auto_taper_taper)
        taper_o1 = taper_.ports[0].name
        taper_o2 = taper_.ports[1].name
        ports1_new: list[gf.Port] = []
        ports2_new: list[gf.Port] = []

        for p1, p2 in zip(ports1_, ports2_, strict=False):
            t1 = c << taper_
            t2 = c << taper_
            t1.connect(taper_o1, p1)
            t2.connect(taper_o1, p2)

            ports1_new.append(t1.ports[taper_o2])
            ports2_new.append(t2.ports[taper_o2])

        ports1_ = ports1_new
        ports2_ = ports2_new

        bbox1 = gf.kdb.DBox()
        bbox2 = gf.kdb.DBox()

        for port in ports1_:
            bbox1 += port.dcplx_trans.disp.to_p()

        for port in ports2_:
            bbox2 += port.dcplx_trans.disp.to_p()

        bboxes.append(bbox1)
        bboxes.append(bbox2)

    elif auto_taper:
        bbox1 = gf.kdb.DBox()
        bbox2 = gf.kdb.DBox()
        for port in ports1_:
            bbox1 += port.dcplx_trans.disp.to_p()

        for port in ports2_:
            bbox2 += port.dcplx_trans.disp.to_p()

        ports1_ = add_auto_tapers(
            component, ports1_, cross_section=xs, layer_transitions=layer_transitions
        )
        ports2_ = add_auto_tapers(
            component, ports2_, cross_section=xs, layer_transitions=layer_transitions
        )

        for port in ports1_:
            bbox1 += port.dcplx_trans.disp.to_p()

        for port in ports2_:
            bbox2 += port.dcplx_trans.disp.to_p()

        bboxes.append(bbox1)
        bboxes.append(bbox2)
        # component.shapes(component.kcl.layer(1,0)).insert(bbox)

    if steps and waypoints:
        raise ValueError("Provide only one of steps or waypoints")

    if steps:
        waypoints = []
        x, y = ports1_[0].center
        for d in steps:
            if isinstance(d, dict):
                if not STEP_DIRECTIVES.issuperset(d):
                    raise ValueError(
                        f"Invalid step directives: {list(d.keys() - STEP_DIRECTIVES)}."
                        f"Valid directives are {list(STEP_DIRECTIVES)}"
                    )
                x = d.get("x", x) + d.get("dx", 0)
                y = d.get("y", y) + d.get("dy", 0)
            else:
                raise ValueError(
                    f"Invalid step {d!r}. Each step must be a dict with keys (x, y, dx, dy)."
                )
            waypoints += [(x, y)]  # type: ignore[arg-type]
            if layer_marker:
                marker = component << gf.components.rectangle(
                    size=(10, 10), layer=layer_marker, centered=True
                )
                marker.center = (x, y)

    if waypoints is not None and steps and len(waypoints) < 2:
        x, y = waypoints[-1][0], waypoints[-1][1]  # type: ignore[index]
        x1, y1 = ports1_[0].center
        port2 = ports2_[0]
        x2, y2 = port2.center
        orientation = port2.orientation
        if orientation is not None and int(orientation) in {0, 180}:
            yt = y1 + (y2 - y1) / 3
            ytt = y1 + 2 * (y2 - y1) / 3
            waypoints = [(x, yt), (x, ytt)]
        elif orientation is not None and int(orientation) in {90, 270}:
            xt = x1 + (x2 - x1) / 3
            xtt = x1 + 2 * (x2 - x1) / 3
            waypoints = [(xt, y), (xtt, y)]

    waypoints_: list[kf.kdb.DPoint] | None
    if waypoints is None:
        waypoints_ = None
    elif len(waypoints) == 0:
        waypoints_ = []
    elif not isinstance(waypoints[0], kf.kdb.DPoint):
        waypoints_ = [
            kf.kdb.DPoint(p[0], p[1])  # type: ignore[index]
            for p in waypoints
        ]
    else:
        waypoints_ = [cast("kf.kdb.DPoint", p) for p in waypoints]

    if layer_marker and waypoints_ is not None:
        for p in waypoints_:
            marker = component << gf.components.rectangle(
                size=(10, 10), layer=layer_marker, centered=True
            )
            marker.center = (p.x, p.y)

    if waypoints_ is not None and len(waypoints_) >= 2:
        waypoints_ = _ensure_manhattan_waypoints(waypoints_, start_port=ports1_[0])

    bend90 = (
        bend
        if isinstance(bend, gf.Component)
        else gf.get_component(
            bend, cross_section=cross_section, radius=radius, width=width
        )
    )

    def straight_um(width: float, length: float) -> gf.Component:
        return gf.get_component(
            straight, length=length, cross_section=cross_section, width=width
        )

    if sbend:

        def _sbend(
            c: gf.kf.ProtoTKCell[Any], offset: float, length: float, width: float
        ) -> gf.kf.DInstanceGroup:
            sb = gf.get_component(
                sbend,
                cross_section=cross_section,
                width=width,
                size=(length, offset),
            )
            sb_ref = component << sb
            return gf.kf.DInstanceGroup(insts=[sb_ref], ports=list(sb_ref.ports))

    try:
        route = kf.routing.optical.route_bundle(
            component,
            ports1_,
            ports2_,
            separation=separation,
            straight_factory=straight_um,
            bend90_cell=bend90,
            taper_cell=taper_cell,
            starts=start_straight_length,
            ends=end_straight_length,
            min_straight_taper=min_straight_taper,
            place_port_type=port_type,
            collision_check_layers=[
                c.kcl.layout.get_info(layer) for layer in collision_check_layer_enums
            ]
            if collision_check_layer_enums
            else None,
            on_collision=on_collision,
            allow_width_mismatch=allow_width_mismatch,
            allow_layer_mismatch=allow_layer_mismatch,
            allow_type_mismatch=allow_type_mismatch,
            bboxes=list(bboxes or []),
            route_width=width,
            sort_ports=sort_ports,
            waypoints=waypoints_,
            end_angles=end_angles,
            start_angles=start_angles,
            path_length_matching_config=path_length_matching_config,
            sbend_factory=_sbend if sbend else None,
        )
    except Exception as e:
        if "kdb.Trans" in str(e):
            e = ValueError("You need at least 2 waypoints or steps.")
        elif "non-manhattan" in str(e):
            e = ValueError("Waypoints need to be Manhattan (axis-aligned) coordinates.")

        if raise_on_error:
            raise

        gf.logger.error(f"Error in route_bundle: {e}")
        layer_error_path = gf.get_layer_info(gf.CONF.layer_error_path)
        route = kf.routing.electrical.route_bundle(
            component,
            ports1_,
            ports2_,
            separation=separation,
            starts=start_straight_length,
            ends=end_straight_length,
            on_collision=on_collision,
            bboxes=bboxes,
            route_width=width,
            sort_ports=sort_ports,
            end_angles=end_angles,
            start_angles=start_angles,
            place_layer=layer_error_path,
        )

        if waypoints and waypoints_ is not None:
            layer_marker = gf.CONF.layer_error_path
            for p in waypoints_:
                marker = component << gf.components.rectangle(
                    size=(10, 10), layer=layer_marker, centered=True
                )
                marker.center = (p.x, p.y)

    if layer_label:
        for route_i in route:
            c.add_label(
                text=f"{route_i.length:.3f}",
                layer=layer_label,
                position=route_i.instances[0].dcenter,
            )

    return route


route_bundle_electrical = partial(
    route_bundle,
    bend="wire_corner",
    allow_width_mismatch=True,
)



"""Extract netlist from component port connectivity."""

import inspect
import re
import secrets
import warnings
from collections import defaultdict
from collections.abc import Iterable
from hashlib import md5
from itertools import chain, product
from typing import Any, Literal, Protocol, TypeAlias, cast

import kfactory as kf
from natsort import natsorted

import gdsfactory as gf
import gdsfactory.schematic as scm
from gdsfactory.component import Component
from gdsfactory.name import get_instance_name_from_alias as legacy_namer  # noqa: F401
from gdsfactory.serialization import clean_value_json

Instance: TypeAlias = kf.DInstance | kf.VInstance | kf.Instance
ErrorBehavior: TypeAlias = Literal["ignore", "warn", "error"]


class ComponentNamer(Protocol):
    """Protocol for naming components in netlists."""

    def __call__(self, cell: kf.ProtoTKCell[Any]) -> str:
        """Return the component name for the given cell."""
        ...


def factory_namer(cell: kf.ProtoTKCell[Any]) -> str:
    """Names components using their factory name."""
    try:
        return cell.factory_name
    except ValueError:
        return cell.name


def function_namer(cell: kf.ProtoTKCell[Any]) -> str:
    """Names components using their function name, falling back to factory name."""
    try:
        return cell.function_name or cell.factory_name
    except ValueError:
        return cell.name


def cell_namer(cell: kf.ProtoTKCell[Any]) -> str:
    """Names components using their cell name."""
    return cell.name


class InstanceNamer(Protocol):
    """Protocol for naming instances in netlists."""

    def __call__(self, inst: Instance) -> str:
        """Return the instance name for the given instance."""
        ...


class OriginalNamer:
    """Names instances using their original instance name."""

    def __init__(self) -> None:
        self._instance_names: dict[str | None, str] = {}
        self._rev_instance_names: dict[str, str | None] = {}

    def __call__(self, inst: Instance) -> str:
        inst_name = _instname(inst)
        if inst_name in self._instance_names:
            return self._instance_names[inst_name]
        name = self._instance_names[inst_name] = _clean_instname(inst_name)
        self._rev_instance_names[name] = inst_name
        return name


class CountedNamer:
    """Names instances using their component name with numeric suffixes."""

    def __init__(self, component_namer: ComponentNamer) -> None:
        self._component_namer = component_namer
        self._instance_names: dict[str | None, str] = {}
        self._rev_instance_names: dict[str, str | None] = {}

    def __call__(self, inst: Instance) -> str:
        inst_name = _instname(inst)
        if inst_name in self._instance_names:
            return self._instance_names[inst_name]
        compname = _short_component_name(_instcell(inst), self._component_namer)
        name = _instname_from_compname(
            inst, compname, self._instance_names, self._rev_instance_names
        )
        name = self._instance_names[inst_name] = _clean_instname(name)
        self._rev_instance_names[name] = inst_name
        return name


class SmartNamer:
    """Names instances using component name if auto-generated, otherwise instance name."""

    def __init__(self, component_namer: ComponentNamer) -> None:
        self._component_namer = component_namer
        self._instance_names: dict[str | None, str] = {}
        self._rev_instance_names: dict[str, str | None] = {}

    def __call__(self, inst: Instance) -> str:
        inst_name = _instname(inst)
        if inst_name in self._instance_names:
            return self._instance_names[inst_name]
        compname = _short_component_name(_instcell(inst), self._component_namer)
        if inst_name is not None and inst_name.startswith(f"{compname}_"):
            name: str | None = _instname_from_compname(
                inst, compname, self._instance_names, self._rev_instance_names
            )
        else:
            name = inst_name
        cleaned = self._instance_names[inst_name] = _clean_instname(name)
        self._rev_instance_names[cleaned] = inst_name
        return cleaned


class NetlistNamer(Protocol):
    """Protocol for naming cells in recursive netlists."""

    def __call__(self, cell: kf.ProtoTKCell[Any]) -> str:
        """Return the name for the given cell in the netlist."""
        ...


class CountedNetlistNamer:
    """Names cells with counting for uniqueness in recursive netlists.

    Uses component_namer for leaf cells (no instances). For hierarchical cells,
    uses the component name as-is if there are no settings, otherwise uses
    counted naming to distinguish different parameterizations.
    """

    def __init__(self, component_namer: ComponentNamer) -> None:
        self._component_namer = component_namer
        self._cell_names: dict[str, str] = {}  # cell.name -> assigned name
        self._used_names: set[str] = set()

    def __call__(self, cell: kf.ProtoTKCell[Any]) -> str:
        if cell.name in self._cell_names:
            return self._cell_names[cell.name]

        base = self._component_namer(cell)

        if not _has_instances(cell):
            # Leaf cell: use component_namer (typically function_name)
            name = base
        else:
            # Hierarchical cell: use counted naming to avoid collisions
            # between different parameterizations of the same component
            name = self._get_unique_name(base)

        self._cell_names[cell.name] = name
        self._used_names.add(name)
        return name

    def _get_unique_name(self, base: str) -> str:
        # If name is not already used, use it as-is
        if base not in self._used_names:
            return base
        # Otherwise add a numbered suffix
        # If base ends with a number, add underscore
        basename = re.sub("[0-9]*$", "", base)
        if basename != base:
            basename = f"{base}_"
        index = 2
        while (name := f"{basename}{index}") in self._used_names:
            index += 1
        return name


# PortMatcher can return:
# - False: ports don't match
# - True: ports match with no metadata
# - dict: ports match with metadata (stored in net's 'settings' field)
MatchResult: TypeAlias = bool | dict[str, Any]


class PortMatcher(Protocol):
    """Protocol for determining if two ports are connected."""

    def __call__(
        self, port1: kf.DPort | kf.Port, port2: kf.DPort | kf.Port
    ) -> MatchResult:
        """Return match result: False, True, or dict with metadata."""
        ...


class PortCenterMatcher:
    """Matches ports based on center position only."""

    def __init__(self, tolerance_dbu: int = 2) -> None:
        self.tolerance_dbu = tolerance_dbu

    def __call__(self, port1: kf.DPort | kf.Port, port2: kf.DPort | kf.Port) -> bool:
        """Return True if two ports are at the same location (within tolerance)."""
        if port1.port_type != port2.port_type:
            return False
        x1, y1 = port1.to_itype().center
        x2, y2 = port2.to_itype().center
        return abs(x2 - x1) < self.tolerance_dbu and abs(y2 - y1) < self.tolerance_dbu


class SmartPortMatcher:
    """Matches ports based on position, width, and orientation.

    For electrical ports, orientation is not checked.
    For other port types (e.g. optical), ports must face each other (180° apart).
    """

    def __init__(
        self,
        position_tolerance_dbu: int = 2,
        width_tolerance: float = 0.001,
        angle_tolerance: float = 0.01,
    ) -> None:
        self.position_tolerance_dbu = position_tolerance_dbu
        self.width_tolerance = width_tolerance
        self.angle_tolerance = angle_tolerance

    def __call__(self, port1: kf.DPort | kf.Port, port2: kf.DPort | kf.Port) -> bool:
        """Return True if ports match in position, width, and orientation."""
        # Check position
        if port1.port_type != port2.port_type:
            return False
        x1, y1 = port1.to_itype().center
        x2, y2 = port2.to_itype().center
        if (
            abs(x2 - x1) >= self.position_tolerance_dbu
            or abs(y2 - y1) >= self.position_tolerance_dbu
        ):
            return False

        # Check width
        if abs(port1.width - port2.width) > self.width_tolerance:
            return False

        # Skip orientation check for electrical ports
        if port1.port_type == "electrical" or port2.port_type == "electrical":
            return True

        # Check orientation: ports should face each other (180° apart)
        angle_diff = abs(_angle_difference(port1.orientation, port2.orientation))
        if abs(angle_diff - 180) > self.angle_tolerance:  # noqa: SIM103
            return False

        return True


class FlexiblePortMatcher:
    """Matches ports based on position and orientation, recording width mismatches.

    Unlike SmartPortMatcher, this matcher does not reject connections with
    width mismatches. Instead, it returns metadata about the mismatch that
    can be used for post-processing (e.g., inserting interface components).
    """

    def __init__(
        self,
        position_tolerance_dbu: int = 2,
        angle_tolerance: float = 0.01,
    ) -> None:
        self.position_tolerance_dbu = position_tolerance_dbu
        self.angle_tolerance = angle_tolerance

    def __call__(
        self, port1: kf.DPort | kf.Port, port2: kf.DPort | kf.Port
    ) -> MatchResult:
        """Return match result with width mismatch metadata if applicable."""
        # Check port type
        if port1.port_type != port2.port_type:
            return False

        # Check position
        x1, y1 = port1.to_itype().center
        x2, y2 = port2.to_itype().center
        if (
            abs(x2 - x1) >= self.position_tolerance_dbu
            or abs(y2 - y1) >= self.position_tolerance_dbu
        ):
            return False

        # Skip orientation check for electrical ports
        if port1.port_type != "electrical":
            # Check orientation: ports should face each other (180° apart)
            angle_diff = abs(_angle_difference(port1.orientation, port2.orientation))
            if abs(angle_diff - 180) > self.angle_tolerance:
                return False

        # Check for width mismatch - record but don't reject
        width_diff = abs(port1.width - port2.width)
        if width_diff > 0.001:  # small tolerance for floating point
            return {
                "width1": port1.width,
                "width2": port2.width,
            }

        return True


def _angle_difference(angle1: float, angle2: float) -> float:
    """Return the difference between two angles, normalized to [-180, 180]."""
    diff = angle2 - angle1
    while diff < -180:
        diff += 360
    while diff > 180:
        diff -= 360
    return diff


def _flip_port(port: kf.DPort | kf.Port) -> kf.DPort:
    """Return a copy of the port with orientation flipped by 180°."""
    return kf.DPort(
        name=port.name,
        center=(port.x, port.y),
        orientation=port.orientation + 180,
        width=port.width,
        layer=port.layer,
        port_type=port.port_type,
    )


_default_port_matcher = SmartPortMatcher()


def get_netlist(
    cell: kf.ProtoTKCell[Any],
    *,
    on_multi_connect: ErrorBehavior = "error",
    on_dangling_port: ErrorBehavior = "warn",
    instance_namer: InstanceNamer | None = None,
    component_namer: ComponentNamer = function_namer,
    port_matcher: PortMatcher | None = None,
) -> dict[str, Any]:
    """Extract netlist from a cell's port connectivity.

    Args:
        cell: The cell to extract the netlist from.
        on_multi_connect: What to do when more than two ports overlap.
            "ignore": silently allow, "warn": allow with warning, "error": raise.
        on_dangling_port: What to do when an instance port is not connected.
            "ignore": silently allow, "warn": allow with warning, "error": raise.
        instance_namer: Callable to name instances.
            Defaults to SmartNamer(component_namer).
        component_namer: Callable to name components.
            Defaults to function_namer.
        port_matcher: Callable to determine if two ports are connected.
            Defaults to SmartPortMatcher().

    Returns:
        A dictionary containing instances, placements, ports, and nets.
    """
    recnet: dict[str, dict[str, Any]] = {}
    _insert_netlist(
        recnet,
        cell,
        on_multi_connect,
        on_dangling_port,
        instance_namer or SmartNamer(component_namer),
        component_namer,
        component_namer,  # netlist_namer: use component_namer for non-recursive
        port_matcher or _default_port_matcher,
        recursive=False,
    )
    return cast(dict[str, Any], clean_value_json(recnet[next(iter(recnet))]))


def get_netlist_recursive(
    cell: kf.ProtoTKCell[Any],
    *,
    on_multi_connect: ErrorBehavior = "error",
    on_dangling_port: ErrorBehavior = "warn",
    instance_namer: InstanceNamer | None = None,
    component_namer: ComponentNamer = function_namer,
    netlist_namer: NetlistNamer | None = None,
    port_matcher: PortMatcher | None = None,
) -> dict[str, Any]:
    """Extract netlists recursively from a cell and all its subcells.

    Args:
        cell: The cell to extract the netlist from.
        on_multi_connect: What to do when more than two ports overlap.
            "ignore": silently allow, "warn": allow with warning, "error": raise.
        on_dangling_port: What to do when an instance port is not connected.
            "ignore": silently allow, "warn": allow with warning, "error": raise.
        instance_namer: Callable to name instances.
            Defaults to SmartNamer(component_namer).
        component_namer: Callable to name components in instance dicts.
            Defaults to function_namer.
        netlist_namer: Callable to name cells in the recursive netlist.
            Defaults to CountedNetlistNamer(component_namer).
        port_matcher: Callable to determine if two ports are connected.
            Defaults to SmartPortMatcher().

    Returns:
        A dictionary mapping cell names to their netlists.
    """
    recnet: dict[str, Any] = {}
    _insert_netlist(
        recnet,
        cell,
        on_multi_connect,
        on_dangling_port,
        instance_namer or SmartNamer(component_namer),
        component_namer,
        netlist_namer or CountedNetlistNamer(component_namer),
        port_matcher or _default_port_matcher,
        recursive=True,
    )
    return cast(dict[str, dict[str, Any]], clean_value_json(recnet))


def _insert_netlist(
    recnet: dict[str, Any],
    cell: kf.ProtoTKCell[Any],
    on_multi_connect: ErrorBehavior,
    on_dangling_port: ErrorBehavior,
    instance_namer: InstanceNamer,
    component_namer: ComponentNamer,
    netlist_namer: NetlistNamer,
    port_matcher: PortMatcher,
    recursive: bool,
) -> None:
    cell_name = netlist_namer(cell)
    if cell_name in recnet:
        return
    net = recnet[cell_name] = {
        "instances": {},
        "placements": {},
        "ports": {},
        "nets": {},
    }
    _all_ports: dict[str, kf.DPort | kf.Port] = {}
    for _inst in sorted(
        chain(cell.insts, cell.vinsts), key=lambda i: getattr(i, "name", "") or ""
    ):
        inst = cast(Instance, _inst)
        inst_cell = _instcell(inst)
        inst_name = instance_namer(inst)
        if _is_pure_vinst(inst):
            if _is_array_inst(inst):
                msg = (
                    "Cannot export netlist: virtual array instances are not supported."
                )
                raise ValueError(msg)
            array = None
            virtual = True
            transform = inst.dcplx_trans
            _all_ports.update({f"{inst_name},{p.name}": p for p in inst.ports})
        elif _is_flattened_vinst(inst):
            if _is_array_inst(inst):
                msg = (
                    "Cannot export netlist: virtual array instances are not supported."
                )
            if hasattr(inst, "to_dtype"):  # always True - just making mypy happy
                inst = inst.to_dtype()
            array = None
            virtual = True
            transform = inst.dcplx_trans * inst.cell.vtrans  # type: ignore[union-attr]
            _all_ports.update({f"{inst_name},{p.name}": p for p in inst.ports})
        elif _is_array_inst(inst):
            if hasattr(inst, "to_dtype"):  # always True - just making mypy happy
                inst = inst.to_dtype()
            array = _get_array_config(inst)
            virtual = False
            transform = inst.dcplx_trans
            _all_ports.update(
                {
                    f"{inst_name}<{a}.{b}>,{p.name}": inst.ports[p.name, a, b]
                    for p, a, b in product(
                        inst_cell.ports, range(inst.na), range(inst.nb)
                    )
                }
            )
        else:
            if hasattr(inst, "to_dtype"):  # always True - just making mypy happy
                inst = inst.to_dtype()
            array = None
            virtual = False
            transform = inst.dcplx_trans
            _all_ports.update({f"{inst_name},{p.name}": p for p in inst.ports})

        net["instances"][inst_name] = _dump_instance(
            {
                "component": component_namer(inst_cell),
                "array": array,
                "settings": inst_cell.settings.model_dump(),
                "info": inst_cell.info.model_dump(),
                "virtual": virtual,
            }
        )
        net["placements"][inst_name] = {
            "x": transform.disp.x,
            "y": transform.disp.y,
            "rotation": transform.angle,
            "mirror": transform.mirror,
        }

        if recursive and _has_instances(inst_cell):
            net["instances"][inst_name]["component"] = netlist_namer(inst_cell)
            _insert_netlist(
                recnet,
                inst_cell,
                on_multi_connect,
                on_dangling_port,
                instance_namer,
                component_namer,
                netlist_namer,
                port_matcher,
                recursive,
            )

    # Add top-level ports
    # we flip them so they face opposite to the instance ports
    # this is necessary as most PortMatchers will assume ports face each other
    for port in cell.ports:
        if port.name is not None:
            _all_ports[port.name] = _flip_port(cast(kf.DPort | kf.Port, port))

    all_nets = _get_nets(_all_ports, on_multi_connect, port_matcher)
    _handle_dangling_ports(_all_ports, all_nets, on_dangling_port)
    net["ports"], net["nets"] = _split_nets_and_ports(all_nets)


def _has_instances(cell: Any) -> bool:
    """Return True if the cell has any instances."""
    return bool(getattr(cell, "insts", False)) or bool(getattr(cell, "vinsts", False))


def _has_non_default_settings(cell: kf.ProtoTKCell[Any]) -> bool:
    """Return True if the cell has settings that differ from factory defaults."""
    settings = cell.settings.model_dump()
    if not settings:
        return False

    # Try to get the factory function to compare defaults
    try:
        factory_name = cell.function_name or cell.factory_name
    except ValueError:
        # No factory name, assume settings are non-default if present
        return bool(settings)

    # Get factory from active PDK
    pdk = gf.get_active_pdk()
    factory = pdk.cells.get(factory_name)
    if factory is None:
        # Factory not found, assume settings are non-default if present
        return bool(settings)

    # Get default parameter values from factory signature
    try:
        sig = inspect.signature(factory)
    except (ValueError, TypeError):
        return bool(settings)

    defaults = {}
    for name, param in sig.parameters.items():
        if param.default is not inspect.Parameter.empty:
            defaults[name] = param.default

    # Check if any setting differs from default
    for key, value in settings.items():
        if value is None:
            continue
        if key not in defaults:
            # Setting not in factory signature, consider it non-default
            return True
        if value != defaults[key]:
            return True

    return False


def _get_array_config(inst: Instance) -> scm.Array:
    kcl = inst.cell.kcl
    # inst.a and inst.b have the instance rotation baked in (but not mirror).
    # The netlist stores the array pitches in the local (pre-rotation) frame,
    # so we need to undo the rotation to get the original pitch vectors.
    trans = inst.dcplx_trans
    inv_rot = kf.kdb.DCplxTrans(1, trans.angle, False, 0, 0).inverted()
    a = inv_rot * kf.kdb.DVector(inst.a.x, inst.a.y)
    b = inv_rot * kf.kdb.DVector(inst.b.x, inst.b.y)
    ax = round(kcl.dbu * a.x, 6)
    ay = round(kcl.dbu * a.y, 6)
    bx = round(kcl.dbu * b.x, 6)
    by = round(kcl.dbu * b.y, 6)
    match (
        ax == 0,
        ay == 0,
        bx == 0,
        by == 0,
    ):
        case (_, True, True, _):
            return scm.OrthogonalGridArray(
                columns=inst.na,
                rows=inst.nb,
                column_pitch=ax,
                row_pitch=by,
            )
        case (True, _, _, True):
            return scm.OrthogonalGridArray(
                columns=inst.nb,
                rows=inst.na,
                column_pitch=bx,
                row_pitch=ay,
            )
    return scm.GridArray(
        num_a=inst.na,
        num_b=inst.nb,
        pitch_a=(ax, ay),
        pitch_b=(bx, by),
    )


def _is_array_inst(inst: Instance) -> bool:
    return getattr(inst, "na", 0) > 1 or getattr(inst, "nb", 0) > 1


def _is_pure_vinst(inst: Instance) -> bool:
    return isinstance(inst, kf.VInstance)


def _is_flattened_vinst(inst: Instance) -> bool:
    return getattr(inst.cell, "vtrans", None) is not None


def _dump_instance(
    instance: dict[str, Any],
    instance_exclude: Iterable[str] = (),
) -> dict[str, Any]:
    instance_exclude = set(instance_exclude)
    dct = {}
    for k, v in instance.items():
        if (
            k in instance_exclude
            or (k == "array" and v is None)
            or (k == "virtual" and v is False)
        ):
            continue
        if hasattr(v, "model_dump"):
            dct[k] = v.model_dump()
        else:
            dct[k] = v
    return dct


def _short_component_name(
    cell: kf.ProtoTKCell[Any], component_namer: ComponentNamer
) -> str:
    """Get short component name, preferring function_name."""
    try:
        return cell.function_name or component_namer(cell)
    except ValueError:
        return component_namer(cell)


def _instname_from_compname(
    inst: Instance,
    compname: str,
    _instance_names: dict[str | None, str],
    _rev_instance_names: dict[str, str | None],
) -> str:
    inst_name = _instname(inst)
    if compname not in _rev_instance_names:
        _instance_names[inst_name] = compname
        _rev_instance_names[compname] = inst_name
        return compname

    # if the compname already ends on a number, we add an underscore.
    basename = re.sub("[0-9]*$", "", compname)
    if basename != compname:
        basename = f"{compname}_"

    index = 2
    while (compname := f"{basename}{index}") in _rev_instance_names:
        index += 1

    _instance_names[inst_name] = compname
    _rev_instance_names[compname] = inst_name
    return compname


def _clean_instname(name: str | None) -> str:
    if name is None:
        return f"unnamed_{secrets.token_hex(4)}"
    replace_map = {" ": "_", "!": "", "?": "", "#": "_", "%": "_", "(": "", ")": "", "*": "_", ",": "_", "-": "m", ".": "p", "/": "_", ":": "_", "=": "", "@": "_", "[": "", "]": "", "{": "", "}": "", "$": ""}  # fmt: skip
    for k, v in replace_map.items():
        name = name.replace(k, v)
    name = re.sub("[^a-zA-Z0-9]", "_", name)
    if name[0] in "0123456789":
        name = f"_{name}"
    return name


def _handle_multi_connect(
    matched_pairs: dict[tuple[str, str], dict[str, Any] | None],
    all_ports: dict[str, kf.DPort | kf.Port],
    on_multi_connect: ErrorBehavior,
) -> None:
    """Check for multiple connections at same location and warn/error as configured."""
    if on_multi_connect == "ignore":
        return

    # Group pairs by shared ports to detect multi-port overlaps
    port_connections: dict[str, set[str]] = defaultdict(set)
    for p1, p2 in matched_pairs:
        port_connections[p1].add(p2)
        port_connections[p2].add(p1)

    for port, connected in port_connections.items():
        if len(connected) > 1:
            p = all_ports[port]
            x, y = p.x, p.y
            msg = f"More than two ports overlapping at ({x:.3f}, {y:.3f}), {port}: {connected | {port}}."
            if on_multi_connect == "error":
                raise ValueError(msg)
            warnings.warn(msg, stacklevel=5)


def _get_nets(
    all_ports: dict[str, kf.DPort | kf.Port],
    on_multi_connect: ErrorBehavior,
    port_matcher: PortMatcher,
) -> list[dict[str, Any]]:
    """Extract connections between ports.

    Uses spatial indexing to bucket ports by position for ~O(n) matching
    instead of O(n²).

    Returns:
        List of net dicts with keys 'p1', 'p2', and optionally 'settings'.
    """
    _matched_pairs: dict[tuple[str, str], dict[str, Any] | None] = {}

    # Spatial index: bucket ports by position (bucket size in dbu)
    _BUCKET = 5
    buckets: dict[tuple[int, int], list[tuple[str, kf.DPort | kf.Port]]] = defaultdict(
        list
    )
    for pname, p in all_ports.items():
        cx, cy = p.to_itype().center
        buckets[cx // _BUCKET, cy // _BUCKET].append((pname, p))

    # Compare ports within same and neighboring buckets only
    for (bx, by), ports in buckets.items():
        candidates: list[tuple[str, kf.DPort | kf.Port]] = []
        for dx in (-1, 0, 1):
            for dy in (-1, 0, 1):
                nbr = buckets.get((bx + dx, by + dy))
                if nbr is not None:
                    candidates.extend(nbr)
        for pname, p in ports:
            for qname, q in candidates:
                if pname >= qname:
                    continue
                result = port_matcher(p, q)
                if result is not False:
                    if isinstance(result, dict):
                        _matched_pairs[pname, qname] = result
                    else:
                        _matched_pairs[pname, qname] = None

    _handle_multi_connect(_matched_pairs, all_ports, on_multi_connect)

    # Build list of nets
    nets: list[dict[str, Any]] = []
    for (p1, p2), settings in _matched_pairs.items():
        net: dict[str, Any] = {"p1": p1, "p2": p2}
        if settings is not None:
            net["settings"] = settings
        nets.append(net)
    return nets


def _split_nets_and_ports(
    nets: list[dict[str, Any]],
) -> tuple[dict[str, str], list[dict[str, Any]]]:
    """Split nets into top-level port mappings and instance-to-instance nets.

    Top-level ports never have a ',' in their name, whereas instance ports
    are formatted as '{instance_name},{port_name}'.

    Note: Settings from nets involving top-level ports are discarded since
    the ports dict only maps port names to instance ports.
    """
    ports: dict[str, str] = {}
    instance_nets: list[dict[str, Any]] = []
    for net in nets:
        p1, p2 = net["p1"], net["p2"]
        if "," not in p1:
            ports[p1] = p2
        elif "," not in p2:
            ports[p2] = p1
        else:
            p1 = net["p1"]
            p2 = net["p2"]
            meta = {k: v for k, v in net.items() if k != "p1" and k != "p2"}
            if p1 < p2:
                instance_nets.append({"p1": p1, "p2": p2, **meta})
            else:
                instance_nets.append({"p1": p2, "p2": p1, **meta})
    ports = {k: ports[k] for k in natsorted(ports)}
    nets = natsorted(instance_nets, key=lambda n: f"{n['p1']}:{n['p2']}")
    return ports, nets


def _handle_dangling_ports(
    all_ports: dict[str, kf.DPort | kf.Port],
    nets: list[dict[str, Any]],
    on_dangling_port: ErrorBehavior,
) -> None:
    """Check for unconnected instance ports and warn/error as configured.

    Top-level ports (without ',' in name) are not considered dangling.
    """
    if on_dangling_port == "ignore":
        return

    # Collect all connected ports
    connected: set[str] = set()
    for net in nets:
        connected.add(net["p1"])
        connected.add(net["p2"])

    # Find dangling instance ports (exclude top-level ports)
    dangling = [p for p in all_ports if "," in p and p not in connected]

    if dangling:
        msg = f"Unconnected ports: {dangling}"
        if on_dangling_port == "error":
            raise ValueError(msg)
        warnings.warn(msg, stacklevel=4)


@gf.cell
def _sample_circuit() -> Component:
    c = Component()
    ring = c.add_ref(gf.c.ring_single(), name="ring").move((100, 0))
    mzi = c.add_ref_off_grid(gf.c.mzi()).rotate(33).move((0, 20))
    mzi.name = "mzi"
    s1 = c.add_ref_off_grid(gf.c.bend_euler_all_angle(angle=90 - 33)).connect(  # type: ignore[arg-type]
        "o1", mzi["o1"]
    )
    s1.name = "s1"
    s2 = c.add_ref_off_grid(gf.c.bend_euler_all_angle(angle=33)).connect(  # type: ignore[arg-type]
        "o2", mzi["o2"]
    )
    s2.name = "s2"
    gf.routing.route_bundle(c, [s1["o2"]], [ring["o1"]], cross_section="strip")
    arr = c.add_ref(gf.c.straight(), columns=1, rows=4, row_pitch=30).move((150, 0))
    arr.name = "arr"
    gf.routing.route_bundle(
        c,
        [arr["o1", 0, 0], arr["o1", 0, 3]],
        [ring["o2"], s2["o1"]],
        cross_section="strip",
    )
    gf.routing.route_bundle(
        c, [arr["o1", 0, 1]], [arr["o1", 0, 2]], cross_section="strip"
    )
    for i, p in enumerate(list(arr.ports)[1::2]):
        c.add_port(name=f"o{i + 1}", port=p)
    return c


def _width_mismatch_circuit() -> Component:
    """Create a simple circuit with a width mismatch for testing."""
    c = Component()
    s1 = c.add_ref(gf.c.straight(length=10, width=0.5), name="s1")  # noqa: F841
    s2 = c.add_ref(gf.c.straight(length=10, width=0.6), name="s2")
    s2.move((10, 0))  # Position s2 so its o1 aligns with s1's o2
    return c


def _instname(inst: Instance) -> str:
    cell = _instcell(inst)
    h = md5(
        repr(
            (
                cell.name,
                (
                    inst.trans.disp.x,
                    inst.trans.disp.y,
                    inst.trans.angle,
                    inst.trans.mirror,
                ),
                inst.a,
                inst.b,
                inst.na,
                inst.nb,
            )
        ).encode()
    ).hexdigest()[:8]
    return inst.name or f"{cell.name}__{h}"


def _instcell(inst: Instance) -> kf.ProtoTKCell[Any]:
    if _is_flattened_vinst(inst):
        inst_cell = gf.get_component(
            str(inst.cell.function_name or inst.cell.factory_name),
            **inst.cell.settings.model_dump(),
        )
        return cast(kf.ProtoTKCell[Any], inst_cell)
    return cast(kf.ProtoTKCell[Any], inst.cell)


if __name__ == "__main__":
    import gdsfactory as gf
    from gdsfactory.gpdk import PDK

    PDK.activate()

    # Test sample circuit
    c = _sample_circuit()
    netlist = c.get_netlist()
    # gf.clear_cache()
    c2 = gf.read.from_yaml(netlist)
    c2.show()
